/*******************************************************************************
 * application.c
 * Network server that listens and forks a session for each connection
 *
 * June 26, 2004
 *
 * Michael Siegenthaler
 * www.msigi.net
 *
 ******************************************************************************/
#define _GNU_SOURCE
#include <sys/types.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include <signal.h>
#include <sys/wait.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "application.h"
#include "log.h"
#include "daemon.h"
#include "loadconfig.h"
#include "session.h"

//#define APPLICATION_DEBUG 1
#ifdef APPLICATION_DEBUG
FILE* err;
#endif

#define BACKLOG 10

// Private function declarations
void sigchild_handler(int s);

/*******************************************************************************
 * Main entry point for application
 ******************************************************************************/
void application (void) {

  // Variables
  char* str = 0;

  // Configuration file variables
  int* configIntP = 0;
  unsigned short int listenPort;

  // Network related variables
  int listen_sd, session_sd;
  struct sockaddr_in local_addr;
  struct sockaddr_in remote_addr;
  int sin_size;
  struct sigaction sa;
  int yes = 1;
  
  // Informational message
  logMessage("Starting application");

  // Determine port to listen on
  if ( !(configIntP = getIntegerByName("listenport")) ) {
    logMessage("FATAL: Cannot get listenport from config");
    exit(EXIT_FAILURE);
  }
  listenPort = (unsigned short int)*configIntP;

  // Create listening socket
  if ( (listen_sd = socket(AF_INET, SOCK_STREAM, 0)) ==  -1) {
    logMessage("FATAL: Cannot create listening socket");
    exit(EXIT_FAILURE);
  }

  // Set listening socket options
  if ( setsockopt(listen_sd, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(int)) == -1 ) {
    logMessage("FATAL: Cannot set listening socket options");
    exit(EXIT_FAILURE);
  }

  // Set up address variables
  local_addr.sin_family = AF_INET;
  local_addr.sin_port = htons(listenPort);
  local_addr.sin_addr.s_addr = INADDR_ANY;
  memset(&(local_addr.sin_zero), '\0', 8);
  memset(&remote_addr, '\0', sizeof(struct sockaddr_in));

  // Bind listening socket
  if ( bind(listen_sd, (struct sockaddr *)&local_addr,
             sizeof(struct sockaddr)) == -1 ) {
    asprintf(&str, "FATAL: Cannot bind listening socket to port %d", (int)listenPort);
    dLogMessage(str);
    exit(EXIT_FAILURE);
  }
  asprintf(&str, "Listening on port %d", (int)listenPort);
  dLogMessage(str);

  // After we have bound to the port, we can drop root privileges
  dropPrivileges();

  // Listen on listening socket
  if (listen(listen_sd, BACKLOG) == -1) {
    logMessage("FATAL: Cannot listen on listening socket");
    exit(EXIT_FAILURE);
  }

  // Reap all dead processes
  sa.sa_handler = sigchild_handler;
  sigemptyset(&sa.sa_mask);
  sa.sa_flags = SA_RESTART;
  if (sigaction(SIGCHLD, &sa, NULL) == -1) {
    logMessage("FATAL: Cannot install SIGCHLD handler");
    exit(EXIT_FAILURE);
  }


#ifdef APPLICATION_DEBUG
  err = getLogFileStream();
  fprintf(err,"about to enter accept loop\n");
#endif

  // Loop and accept connections
  while (1) {

#ifdef APPLICATION_DEBUG
    fprintf(err,"begin accept loop\n");
#endif

    sin_size = sizeof(struct sockaddr_in);
    if ( (session_sd = accept(listen_sd, (struct sockaddr *)&remote_addr,
                               &sin_size)) == -1 ) {
      logMessage("ERROR: Could not accept connection");
      continue;
    }

#ifdef APPLICATION_DEBUG
    fprintf(err, "accepted connection, session_sd=%d\n", session_sd);
#endif

    // Create a new process to handle the session
    if (!fork()) {
      // this is the child process
#ifdef APPLICATION_DEBUG
      fprintf(err, "forked (child) pid=%d\n", (int)getpid() );
#endif

      // Informational message
      asprintf(&str, "Accepted connection from %s:%d",
                inet_ntoa(remote_addr.sin_addr),
                (int)ntohs(remote_addr.sin_port));
      dLogMessage(str);

      close(listen_sd);        // not needed by the child

      run_session(session_sd); // run the function that handles a session

      close(session_sd);
      exit(EXIT_SUCCESS);

    }
    close(session_sd); // Parent doesn't need this
#ifdef APPLICATION_DEBUG
      fprintf(err, "forked (parent) pid=%d\n", (int)getpid() );
#endif

  }

}

/*******************************************************************************
 * Reaps all dead child processes
 ******************************************************************************/
void sigchild_handler(int s) {
pid_t pid;
#ifdef APPLICATION_DEBUG
      fprintf(err, "started sigchild_handler()\n");
#endif
  while ( (pid = waitpid(WAIT_ANY, NULL, WNOHANG)) > 0 ) {
#ifdef APPLICATION_DEBUG
      fprintf(err, "wait returned %d\n", (int)pid);
#endif
  }
#ifdef APPLICATION_DEBUG
      fprintf(err, "finished sigchild_handler(), wait returned %d\n", (int)pid);
#endif
}
