/*******************************************************************************
 * daemon.c
 * Converts an interactive process into a daemon
 *
 * June 26, 2004
 *
 * Michael Siegenthaler
 * www.msigi.net
 *
 ******************************************************************************/
#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/stat.h>
#include <signal.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>

#include "log.h"
#include "loadconfig.h"
#include "daemon.h"

// Define this to write user switching information to the log
#define LOG_USER_SWITCH

#define PIDBUFLEN 13

// Module variables
int pidFile_fd;            // file descriptor of pid file
char* pidFileName = 0;     // filename of pid file
pid_t mainProcessPid = 0;  // pid for the main process

// Private function declarations
void createPidFile(void);
void writePidFile(void);
void deletePidFile(void);
void switchCredentials(int newUser, int newGroup);
void sigtermHandlerInstall(void);
void sigtermHandler(int s);

/*******************************************************************************
 * Convert the process into a daemon process
 * The parent exits and the child returns to the caller
 ******************************************************************************/
void makeDaemon() {

  // Variables for daemon operation
  pid_t pid, sid;

  // Temporarily store pid (in case pid file needs to be deleted before fork)
  mainProcessPid = getpid();

  // Get the pid file name from the configuration
  pidFileName = getParamByName("pidfile");
  if (!pidFileName) {
    fprintf(stderr, "FATAL: Pidfile not specified in config.\n");
    exit(EXIT_FAILURE);
  }

  // Create pid file while still attached to a terminal, so the user can
  // be informed of any potential errors
  createPidFile();

  // First fork
  pid = fork();
  if (pid < 0) {
    // fork failed
    fprintf(stderr,"FATAL: fork() failed\n");
    close(pidFile_fd);
    deletePidFile();
    exit(EXIT_FAILURE);
  }

  // Check whether we're in the parent or the child
  if (pid > 0) {
    // In the parent, let's bail
    exit(EXIT_SUCCESS);
  }
  // In the child

  // Record the pid in memory, so that child processes can later be distinguished
  mainProcessPid = getpid();

  // Write the pid into the already-created pid file
  writePidFile();

  // Install a handler for SIGTERM
  sigtermHandlerInstall();

  // The pid file should be deleted on exit
  if ( atexit(deletePidFile) < 0 ) {
    logMessage("FATAL: Cannot register deletePidFile() with atexit()");
    deletePidFile();
    exit(EXIT_FAILURE);
  }

  // Start a new session
  sid = setsid();
  if (sid < 0) {
    // setsid failed
    logMessage("FATAL: setsid() failed");
    exit(EXIT_FAILURE);
  }

  // Make / the current directory
  if ((chdir("/")) < 0) {
    // chdir failed
    logMessage("FATAL: chdir() failed");
    exit(EXIT_FAILURE);
  }

  // Reset the file mode
  umask(0);

  // Close unneeded file descriptors
  close(STDIN_FILENO);
  close(STDOUT_FILENO);
  close(STDERR_FILENO);

  // At this point we are running as a daemon

}

/*******************************************************************************
 * Creates a pid file but does not write to it
 * This must be called while the process is attached to a terminal
 ******************************************************************************/
void createPidFile() {

  // Copy pid file name into a module variable
  //asprintf(&pidFileName, "%s", fName);

  // Create the pid file, failing if it already exists
  pidFile_fd = open(pidFileName, O_WRONLY | O_CREAT | O_EXCL | O_FSYNC, 0644);

  // If pid file creation failed, we have a problem
  if (pidFile_fd < 0) {
    if (errno == EEXIST) {
      fprintf(stderr,"FATAL: PID file already exists\n");
    } else {
      fprintf(stderr,"FATAL: Cannot create PID file\n");
    }
    exit(EXIT_FAILURE);
  }

}

/*******************************************************************************
 * Writes the current pid to the pid file
 * This is called after the process has been converted into a daemon
 ******************************************************************************/
void writePidFile() {

  // Variables
  pid_t myPID;
  char buf[PIDBUFLEN];

  // Get pid as integer and string
  myPID = getpid();
  snprintf(buf, PIDBUFLEN, "%d\n", myPID);

  // Write pid string into the pid file
  if ( write(pidFile_fd, buf, strlen(buf)) < 0 ) {
    logMessage("FATAL: Cannot write pid to file");
    close(pidFile_fd);
    deletePidFile();
    exit(EXIT_FAILURE);
  }

  // Close the pid file
  close(pidFile_fd);

}

/*******************************************************************************
 * Deletes the pid file
 * (this function must be registered with atexit in the daemon (child) process)
 ******************************************************************************/
void deletePidFile(void) {

  pid_t myPID;

  myPID = getpid();

  // delete the pid file only when the main server process exits, and not
  // if any of its children exit
  if ( myPID == mainProcessPid ) {
    if ( unlink(pidFileName) < 0 ) {
      logMessage("ERROR: Cannot delete pid file");
    }
  }

}

/*******************************************************************************
 * Check whether a user and group are specified in the config file, and
 * if so, switch to using those credentials
 ******************************************************************************/
void dropPrivileges() {

  // Variables
  int* intP;
  int switchUser, newUser, newGroup;

  if ( !(intP = getIntegerByName("switchuser")) ) {
    logMessage("FATAL: Cannot get switchuser from config");
    exit(EXIT_FAILURE);
  }
  switchUser = *intP;

  // switchUser is true if we want to switch users, otherwise false
  if (switchUser) {

    // Get the user from the config
    if ( !(intP = getIntegerByName("user")) ) {
      logMessage("FATAL: Cannot get user from config");
      exit(EXIT_FAILURE);
    }
    newUser = *intP;

    // Get the group from the config
    if ( !(intP = getIntegerByName("group")) ) {
      logMessage("FATAL: Cannot get group from config");
      exit(EXIT_FAILURE);
    }
    newGroup = *intP;

    // Switch to using the new credentials
    switchCredentials(newUser, newGroup);

  }

}

/*******************************************************************************
 * Switch user and group
 * Parameters: numerical user and group id to switch to
 ******************************************************************************/
void switchCredentials(int newUser, int newGroup) {

  // Variables
  uid_t newUID;
  gid_t newGID;

#ifdef LOG_USER_SWITCH
  char* str;
#endif

  // Convert passed-in user and group to appropriate types
  newUID = (uid_t) newUser;
  newGID = (gid_t) newGroup;

#ifdef LOG_USER_SWITCH
  // Informational log messages
  asprintf(&str, "Initial: realuid=%d realgid=%d effuid=%d effgid=%d",
            getuid(), getgid(), geteuid(), getegid());
  dLogMessage(str);
  asprintf(&str, "Requested: uid=%d gid=%d", newUser, newGroup);
  dLogMessage(str);
#endif

  // Change ownership of the pid file so that it can be deleted later
  if ( chown(pidFileName, newUID, newGID) < 0 ) {
    logMessage("FATAL: Cannot change ownership of pid file");
    exit(EXIT_FAILURE);
  }

  // Switch the group
  if ( setgid(newGID) < 0 ) {
    logMessage("FATAL: Cannot switch group");
    exit(EXIT_FAILURE);
  }

  // Switch the user
  if ( setuid(newUID) < 0 ) {
    logMessage("FATAL: Cannot switch user");
    exit(EXIT_FAILURE);
  }

#ifdef LOG_USER_SWITCH
  // Informational log message
  asprintf(&str, "Final: realuid=%d realgid=%d effuid=%d effgid=%d",
            getuid(), getgid(), geteuid(), getegid());
  dLogMessage(str);
#endif

}

/*******************************************************************************
 * Installs a signal handler for SIGTERM
 ******************************************************************************/
void sigtermHandlerInstall(void) {

  struct sigaction sa;

  sa.sa_handler = sigtermHandler;
  sigemptyset(&sa.sa_mask);
  sa.sa_flags = SA_RESTART;
  if ( sigaction(SIGTERM, &sa, NULL) == -1 ) {
    logMessage("FATAL: Cannot install SIGTERM handler");
    exit(EXIT_FAILURE);
  }

}

/*******************************************************************************
 * Properly handle SIGTERM
 ******************************************************************************/
void sigtermHandler(int s) {
  logMessage("Termination signal caught by handler");
  exit(EXIT_SUCCESS);
}
