/*******************************************************************************
 * lionet.c
 * Handles line input and output over a network socket
 *
 * December 19, 2005
 *
 * Michael Siegenthaler
 * www.msigi.net
 *
 ******************************************************************************/
#define _GNU_SOURCE
#include <sys/types.h>

#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include <errno.h>
#include <sys/socket.h>
#include <netinet/in.h>

#include "lionet.h"
#include "log.h"

/*******************************************************************************
 * Send line wrapper function
 * Parameters: socket descriptor, null terminated string to send
 * No return value, exits on failure
 ******************************************************************************/
void send_line(int sd, char* ibuf) {
  int rv;
  rv = _send_line(sd, ibuf);
  if (rv < 0) {
    char* str;
    asprintf(&str, "FATAL: _send_line() failed, errno=%d", errno);
    dLogMessage(str);
    exit(EXIT_FAILURE);
  }
} // end send_line()

/*******************************************************************************
 * Send line
 * Parameters: socket descriptor, null terminated string to send
 * Returns 0 on success, -1 on failure
 * Exits if memory cannot be allocated
 ******************************************************************************/
int _send_line(int sd, char* ibuf) {

  // Variables
  int ibuflen;
  int linetermlen;
  int sentlen;
  int rv;
  void* obuf = 0;
  void* obuf2 = 0;

  // Determine lengths
  ibuflen = strlen(ibuf);
  linetermlen = strlen(LIONET_LINETERM);

  // Allocate a temporary buffer
  obuf = malloc(ibuflen + linetermlen);
  if (!obuf) {
    logMessage("FATAL: Cannot allocate memory in send_line()");
    exit(EXIT_FAILURE);
  }

  // Concatenate strings in temporary buffer
  strncpy(obuf, ibuf, ibuflen);
  if (linetermlen != 0) {
    strncpy(obuf+ibuflen, LIONET_LINETERM, linetermlen);
  }

  // Send the contents of the temporary buffer
  obuf2 = obuf;
  sentlen = 0;
  do {
    rv = send(sd, obuf2, ibuflen+linetermlen-sentlen, 0);
    if (rv < 0) {
      // error
      return -1;
    }
    sentlen += rv;
    obuf2 += rv;
  } while (sentlen < ibuflen+linetermlen);

  // Deallocate temporary buffer
  free(obuf);

  return 0;

} // end _send_line()

/*******************************************************************************
 * Send line serial
 * This is like _send_line, except for use with a serial port instead of a
 * network socket
 * Parameters: socket descriptor, null terminated string to send
 * Returns 0 on success, -1 on failure
 * Exits if memory cannot be allocated
 ******************************************************************************/
int _send_line_serial(int sd, char* ibuf) {

  // Variables
  int ibuflen;
  int linetermlen;
  int sentlen;
  int rv;
  void* obuf = 0;
  void* obuf2 = 0;

  // Determine lengths
  ibuflen = strlen(ibuf);
  linetermlen = strlen(LIONET_LINETERM);

  // Allocate a temporary buffer
  obuf = malloc(ibuflen + linetermlen);
  if (!obuf) {
    logMessage("FATAL: Cannot allocate memory in _send_line_serial()");
    exit(EXIT_FAILURE);
  }

  // Concatenate strings in temporary buffer
  strncpy(obuf, ibuf, ibuflen);
  if (linetermlen != 0) {
    strncpy(obuf+ibuflen, LIONET_LINETERM, linetermlen);
  }

  // Send the contents of the temporary buffer
  obuf2 = obuf;
  sentlen = 0;
  do {
    rv = write(sd, obuf2, ibuflen+linetermlen-sentlen);
    if (rv < 0) {
      // error
      return -1;
    }
    sentlen += rv;
    obuf2 += rv;
  } while (sentlen < ibuflen+linetermlen);

  // Deallocate temporary buffer
  free(obuf);

  return 0;

} // end _send_line_serial()

/*******************************************************************************
 * Receive line wrapper function
 * Parameters: socket descriptor, buffer, length of buffer
 * Has no return value, exits on failure
 ******************************************************************************/
void recv_line(int sd, char* buf, int maxlen_in) {
  int rv;
  rv = _recv_line(sd, buf, maxlen_in);
  if (rv == 0) {
    return;
  } else if (rv == 1) {
    exit(EXIT_SUCCESS);
  } else {
    char* str;
    asprintf(&str, "FATAL: _recv_line() failed, errno=%d", errno);
    dLogMessage(str);
    exit(EXIT_FAILURE);
  }
} // end recv_line()

/*******************************************************************************
 * Receive line
 * Parameters: socket descriptor, buffer, length of buffer
 * Returns 0 on success, -1 on failure, 1 if the peer closed the connection
 * Exits if memory cannot be allocated
 * Bugs: If a line is too long, only maxlen_in characters will be returned,
 * and the following call will return the rest
 ******************************************************************************/
int _recv_line(int sd, char* buf, int maxlen_in) {

  static char save[LIONET_BUFLEN];
  static int save_start = 0;
  static int save_end = 0;
  int buf_index = 0;
  int recvlen;

#ifdef LIONET_DEBUG
  FILE* err;
  err = getLogFileStream();
  fprintf(err,"entering recv_line()\n");
#endif

  while (1) {

    // Process saved characters
    while (save_start < save_end) {
#ifdef LIONET_DEBUG
    fprintf(err, "looping with save_start=%d, save_end=%d\n", save_start, save_end);
#endif
      if (save[save_start] == '\n' || save[save_start] == '\r') {
        if (buf_index > 0) {
          // We have a complete line, return it
          buf[buf_index] = '\0';
          save_start++;
          return 0;
        } else {
          // We got a blank line, ignore it
        }
      } else {
        buf[buf_index] = save[save_start];
        buf_index++;
      }
      save_start++;
      if (buf_index == maxlen_in-1) {
        // Return because we are out of buffer space
        buf[buf_index] = '\0';
        return 0;
      }
    }

    // Get more data from the network
    recvlen = recv(sd, save, LIONET_BUFLEN, 0);
#ifdef LIONET_DEBUG
    fprintf(err, "recv returned %d\n", recvlen);
#endif
    if (recvlen == -1) {
      // Failure
      return -1;
    } else if (recvlen == 0) {
      // the connection was closed by the peer
      return 1;
    } else {
      save_start = 0;
      save_end = recvlen;
    }

  }

} // end _recv_line()

/*******************************************************************************
 * Receive line but timeout if nothing received for the specified # of seconds
 * Parameters: socket descriptor, buffer, length of buffer
 * Returns 0 on success, -1 on failure, 1 if the peer closed the connection,
 *         2 on timeout
 * Exits if memory cannot be allocated
 * Bugs: If a line is too long, only maxlen_in characters will be returned,
 * and the following call will return the rest. If a line is received in
 * multiple packets, with a significant amount of time between the, then the
 * operation could take considerably longer than the specified timeout.
 ******************************************************************************/
int _recv_line_timeout(int sd, int timeout, char* buf, int maxlen_in) {

  static char save[LIONET_BUFLEN];
  static int save_start = 0;
  static int save_end = 0;
  int buf_index = 0;
  int recvlen;
  fd_set readfds;
  struct timeval tmout;

#ifdef LIONET_DEBUG
  FILE* err;
  err = getLogFileStream();
  fprintf(err,"entering _recv_line_timeout()\n");
#endif

  while (1) {

    // Process saved characters
    while (save_start < save_end) {
#ifdef LIONET_DEBUG
    fprintf(err, "looping with save_start=%d, save_end=%d\n", save_start, save_end);
#endif
      if (save[save_start] == '\n' || save[save_start] == '\r') {
        if (buf_index > 0) {
          // We have a complete line, return it
          buf[buf_index] = '\0';
          save_start++;
          return 0;
        } else {
          // We got a blank line, ignore it
        }
      } else {
        buf[buf_index] = save[save_start];
        buf_index++;
      }
      save_start++;
      if (buf_index == maxlen_in-1) {
        // Return because we are out of buffer space
        buf[buf_index] = '\0';
        return 0;
      }
    }

    // Set up set for select()
    FD_ZERO(&readfds);    // clear set
    FD_SET(sd, &readfds); // add socket descriptor to set

    // Set up timeout struct
    tmout.tv_sec = timeout;
    tmout.tv_usec = 0;

    select(sd+1, &readfds, NULL, NULL, &tmout);

    if (FD_ISSET(sd, &readfds)) { // Data is ready

      // Get more data from the network
      recvlen = recv(sd, save, LIONET_BUFLEN, 0);
#ifdef LIONET_DEBUG
      fprintf(err, "recv returned %d\n", recvlen);
#endif
      if (recvlen == -1) {
        // Failure
        return -1;
      } else if (recvlen == 0) {
        // the connection was closed by the peer
        return 1;
      } else {
        save_start = 0;
        save_end = recvlen;
      }

    } else {  // Timeout

      return 2;

    }

  }

} // end _recv_line_timeout()

/*******************************************************************************
 * Receive line but timeout if nothing received for the specified # of seconds
 * For use with a serial port, not a network socket
 * Parameters: file descriptor, buffer, length of buffer
 * Returns 0 on success, -1 on failure, 1 if the peer closed the connection,
 *         2 on timeout
 * Exits if memory cannot be allocated
 * Bugs: If a line is too long, only maxlen_in characters will be returned,
 * and the following call will return the rest. If a line is received in
 * multiple packets, with a significant amount of time between the, then the
 * operation could take considerably longer than the specified timeout.
 ******************************************************************************/
int _recv_line_serial(int sd, int timeout, char* buf, int maxlen_in) {

  static char save[LIONET_BUFLEN];
  static int save_start = 0;
  static int save_end = 0;
  int buf_index = 0;
  int recvlen;
  fd_set readfds;
  struct timeval tmout;

#ifdef LIONET_DEBUG
  FILE* err;
  err = getLogFileStream();
  fprintf(err,"entering _recv_line_timeout()\n");
#endif

  while (1) {

    // Process saved characters
    while (save_start < save_end) {
#ifdef LIONET_DEBUG
    fprintf(err, "looping with save_start=%d, save_end=%d\n", save_start, save_end);
#endif
      if (save[save_start] == '\n' || save[save_start] == '\r') {
        if (buf_index > 0) {
          // We have a complete line, return it
          buf[buf_index] = '\0';
          save_start++;
          return 0;
        } else {
          // We got a blank line, ignore it
        }
      } else {
        buf[buf_index] = save[save_start];
        buf_index++;
      }
      save_start++;
      if (buf_index == maxlen_in-1) {
        // Return because we are out of buffer space
        buf[buf_index] = '\0';
        return 0;
      }
    }

    // Set up set for select()
    FD_ZERO(&readfds);    // clear set
    FD_SET(sd, &readfds); // add socket descriptor to set

    // Set up timeout struct
    tmout.tv_sec = timeout;
    tmout.tv_usec = 0;

    select(sd+1, &readfds, NULL, NULL, &tmout);

    if (FD_ISSET(sd, &readfds)) { // Data is ready

      // Get more data from the serial port
      recvlen = read(sd, save, LIONET_BUFLEN);
#ifdef LIONET_DEBUG
      fprintf(err, "read returned %d\n", recvlen);
#endif
      if (recvlen == -1) {
        // Failure
        return -1;
      } else if (recvlen == 0) {
        // the connection was closed by the peer
        return 1;
      } else {
        save_start = 0;
        save_end = recvlen;
      }

    } else {  // Timeout

      return 2;

    }

  }

} // end _recv_line_serial()
