/*******************************************************************************
 * loadconfig.c
 * Configuration file loader and parser
 *
 * June 26, 2004
 *
 * Michael Siegenthaler
 * www.msigi.net
 *
 ******************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "loadconfig.h"

#define BUFLEN 1024
#define MAX_CONFIG_FILE_LINES 1024

#define READING_NEW_LINE 0
#define READING_NAME 1
#define READING_VALUE 2
#define READING_COMMENT 3

char* names[MAX_CONFIG_FILE_LINES];
char* values[MAX_CONFIG_FILE_LINES];

// Private function declarations
void handleCharacter(char c);
void initializeArrays(void);

/*******************************************************************************
 * Loads the config file into memory
 * Parameter: Filename for the config file
 * Returns 0 on success, or -1 on failure
 ******************************************************************************/
int loadConfig(char* fileName) {

  FILE* configFileStream = 0;  // file descriptor
  int character;  // character read as integer
  char c;  // character read

  // Make sure the arrays are zeroed before returning
  initializeArrays();

  // If not filename was supplied, return
  if (!fileName) { return -1; }

  // Open the config file for reading, return on error
  configFileStream = fopen(fileName,"r");
  if (!configFileStream) { return -1; }

  // read file using getc
  while (1) {
    character = getc(configFileStream);
    if (character == EOF) {
      break;
    } else {
      c = (char) character;
      handleCharacter(c);
    }
  }

  // Close config file
  fclose(configFileStream);

  return 0;
}

/*******************************************************************************
 * Returns a parameter that has been read from the config file
 * Parameter: Name of parameter to retrieve
 * Returns a pointer to a string containing the value, or a null pointer if
 * the requested parameter is not found
 ******************************************************************************/
char* getParamByName(char* paramName) {

  int i;

  for ( i=0; i<MAX_CONFIG_FILE_LINES; i++) {
    if (!names[i]) {
      return 0;
    }
    if (!strcmp(paramName,names[i])) {
      return values[i];
    }
  }

  return 0;

}

/*******************************************************************************
 * Returns an integer that has been read from the config file
 * Parameters: Name of integer to retrieve
 * Returns a pointer to a statically allocated integer, or a null pointer
 * upon failure
 ******************************************************************************/
int* getIntegerByName(char* paramName) {

  int i;
  int rv;
  static int theInteger = 0;

  for ( i=0; i<MAX_CONFIG_FILE_LINES; i++) {
    if (names[i] == 0) {
      // reached end of array without finding requested parameter
      return 0;
    }
    if (!strcmp(paramName,names[i])) {
      //return values[i];
      if (values[i] == 0) {
        // no value is defined for this parameter
        return 0;
      }
      rv = sscanf(values[i], "%d", &theInteger);
      if (rv != 1) {
        // Unable to parse the parameter
        return 0;
      }
      return &theInteger;
    }
  }

  return 0;

}

/*******************************************************************************
 * Deals with a single character that was read (private)
 * Parameter: Character
 ******************************************************************************/
void handleCharacter(char c) {

  static int state = READING_NEW_LINE;
  static int bufWriteIndex;
  static char buf[BUFLEN];
  static int paramListIndex = 0;

  if (bufWriteIndex == BUFLEN) {
    fprintf(stderr,"A token in the configuration file is too long.\n");
    exit(EXIT_FAILURE);
  }
  if (paramListIndex == MAX_CONFIG_FILE_LINES) {
    fprintf(stderr,"The configuration file contains too many lines.\n");
    exit(EXIT_FAILURE);
  }

  if (state == READING_NEW_LINE) {
    // expect an alphanumeric
    bufWriteIndex = 0;
    if (isalpha(c) || isdigit(c)) {
      buf[bufWriteIndex] = c;
      bufWriteIndex++;
      state = READING_NAME;
    } else if (c == '#') {
      state = READING_COMMENT;
    }
  } else if (state == READING_NAME) {
    // expect an alphanumeric or an equals sign
    if (isalpha(c) || isdigit(c)) {
      buf[bufWriteIndex] = c;
      bufWriteIndex++;
    } else if (c == '=') {
      // save the name that has been read
      names[paramListIndex] = malloc(bufWriteIndex+1);
      memcpy(names[paramListIndex], buf, bufWriteIndex);
      names[paramListIndex][bufWriteIndex] = '\0';
      bufWriteIndex = 0;
      state = READING_VALUE;
    } else if (c == '\n') {
      state = READING_NEW_LINE;
      bufWriteIndex = 0;
    }
  } else if (state == READING_VALUE) {
    // accept any character, but handle a newline specially
    if (c == '\n' || c == '\r') {
      state = READING_NEW_LINE;
      if (bufWriteIndex) {
        // save the value that has been read
        values[paramListIndex] = malloc(bufWriteIndex+1);
        memcpy(values[paramListIndex], buf, bufWriteIndex);
        values[paramListIndex][bufWriteIndex] = '\0';
        paramListIndex++;
        bufWriteIndex = 0;
      } else {
        free(names[paramListIndex]);
        names[paramListIndex] = 0;
      }
    } else {
      // save the received character
      buf[bufWriteIndex] = c;
      bufWriteIndex++;
    }
  } else if (state == READING_COMMENT) {
    if (c == '\n' || c == '\r') {
      state = READING_NEW_LINE;
    }
  }

}

/*******************************************************************************
 * Initializes the names and value arrays to contain zeros (private)
 ******************************************************************************/
void initializeArrays() {

  int i;

  for (i=0; i<MAX_CONFIG_FILE_LINES; i++) {
    names[i] = 0;
    values[i] = 0;
  }

}

/*******************************************************************************
 * Prints out the contents of the configfile for debugging purposes
 ******************************************************************************/
void printArrayDebug() {

  int i;

  for ( i=0; i<MAX_CONFIG_FILE_LINES; i++) {
    if (names[i] != 0) {
      printf("%i:%s:",i,names[i]);
      if (values[i] != 0) {
        printf("%s",values[i]);
      }
      printf("\n");
    }
  }

}
