/*******************************************************************************
 * serialcomm.c
 * Serial Instrument Driver Generic Functions
 *
 * December 19, 2005
 *
 * Michael Siegenthaler
 * www.msigi.net
 *
 ******************************************************************************/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <fcntl.h>
#include <termios.h>

#include "serialcomm.h"
#include "session.h"
#include "lionet.h"
//#include "log.h"

int serial_timeout_val = 10; // Default timeout (in seconds)

// Routing table
char * serial_routing_table_device[SERIAL_MAX_INSTRUMENTS];
unsigned short int serial_routing_table_flowcontrol[SERIAL_MAX_INSTRUMENTS];
int serial_routing_table_baudrate[SERIAL_MAX_INSTRUMENTS];

// Terminal settings
struct termios oldtio, newtio;

/*******************************************************************************
 * Send an instrument command
 ******************************************************************************/
int serial_inst_write (int inst, char *cmd)
{

  int rv;
  rv = _send_line_serial(inst, cmd);
  if (rv < 0) {
    return -1;
  }
  return 0;

} // end serial_inst_write()


/*******************************************************************************
 * Read from the instrument
 ******************************************************************************/
int serial_inst_read (int inst, char *buf, int buflen)
{

  int rv;
  rv = _recv_line_serial(inst, serial_timeout_val, buf, buflen);
  if (rv != 0) return -1;
  return 0;

} // end serial_inst_read()

/*******************************************************************************
 * Initialize instrument
 * Returns an instrument handle on success
 ******************************************************************************/
int serial_inst_init (int inst_addr)
{
  int inst_fd, inst_idx;

  // Make sure address is valid
  inst_idx = inst_addr - BACKEND_ADDRESS_START_SERIAL;
  if (inst_idx < 0 || inst_idx >= SERIAL_MAX_INSTRUMENTS) {
    return -1;
  }

  // Make sure we have a route to this serial port
  if (!serial_routing_table_device[inst_idx]) return -1;

  inst_fd = open(serial_routing_table_device[inst_idx], O_RDWR | O_NOCTTY );
  if (inst_fd < 0) {
    return -1;
  }

  tcgetattr(inst_fd, &oldtio);
  memset(&newtio, 1, sizeof(newtio));

  newtio.c_cflag = CS8 | CLOCAL | CREAD;
  newtio.c_iflag = IGNPAR | ICRNL;
  newtio.c_oflag = 0;
  newtio.c_lflag = ICANON; // canonical input

  if (serial_routing_table_flowcontrol[inst_idx] == SERIAL_FLOWCONTROL_HARDWARE) {
    newtio.c_cflag |= CRTSCTS;
  } else if (serial_routing_table_flowcontrol[inst_idx] == SERIAL_FLOWCONTROL_SOFTWARE) {
    newtio.c_iflag |= IXON | IXOFF;
  }

  serial_inst_set_speed(serial_routing_table_baudrate[inst_idx]);

  newtio.c_cc[VINTR]    = 0;     /* Ctrl-c */
  newtio.c_cc[VQUIT]    = 0;     /* Ctrl-\ */
  newtio.c_cc[VERASE]   = 0;     /* del */
  newtio.c_cc[VKILL]    = 0;     /* @ */
  newtio.c_cc[VEOF]     = 4;     /* Ctrl-d */
  newtio.c_cc[VTIME]    = 0;     /* inter-character timer unused */
  newtio.c_cc[VMIN]     = 1;     /* blocking read until 1 character arrives */
  //newtio.c_cc[VSWTC]    = 0;     /* '\0' */
  newtio.c_cc[VSTART]   = 0;     /* Ctrl-q */
  newtio.c_cc[VSTOP]    = 0;     /* Ctrl-s */
  newtio.c_cc[VSUSP]    = 0;     /* Ctrl-z */
  newtio.c_cc[VEOL]     = 0;     /* '\0' */
  newtio.c_cc[VREPRINT] = 0;     /* Ctrl-r */
  newtio.c_cc[VDISCARD] = 0;     /* Ctrl-u */
  newtio.c_cc[VWERASE]  = 0;     /* Ctrl-w */
  newtio.c_cc[VLNEXT]   = 0;     /* Ctrl-v */
  newtio.c_cc[VEOL2]    = 0;     /* '\0' */

  tcflush(inst_fd, TCIFLUSH);
  tcsetattr(inst_fd,TCSANOW,&newtio);

  return inst_fd;
} // end serial_inst_init()


/*******************************************************************************
 * Close communications channel to instrument
 ******************************************************************************/
int serial_inst_term (int inst)
{
  tcsetattr(inst, TCSANOW, &oldtio);
  close(inst);
  return 0;
} // end tcp_inst_term()


/*******************************************************************************
 * Set timeout value
 ******************************************************************************/
int serial_set_timeout_val (int tmoutval)
{
  if (tmoutval < SERIAL_TIMEOUT_MIN || tmoutval > SERIAL_TIMEOUT_MAX) return -1;
  serial_timeout_val = tmoutval;
  return 0;
} // end serial_set_timout()

/*******************************************************************************
 * Load instrument routes
 * Return 0 on success, -1 on failure
 ******************************************************************************/
int serial_load_inst_routes (char * serial_config_file)
{

  FILE* routeFileStream = 0;
  char line[1024];
  unsigned int inst_id;
  int baudrate;
  char device[100];
  char flowcontrol[100];
  int i;
  size_t len;

  for (i=0;i<SERIAL_MAX_INSTRUMENTS;i++) {
    serial_routing_table_device[i] = (char*)0;
    serial_routing_table_flowcontrol[i] = (unsigned short int)0;
    serial_routing_table_baudrate[i] = 0;
  }

  routeFileStream = fopen(serial_config_file, "r");
  if (!routeFileStream) {
    return -1;
  }

  while ( fgets(line, 1024, routeFileStream) != 0 ) {

    memset(device, 0, 100);

    if ( line[0] == '#' || line[0] == '\n' || line[0] == '\r' ) continue;
    if ( sscanf(line, "%d%99s%d%99s", &inst_id, device, &baudrate, flowcontrol) != 4 ) {
      fclose(routeFileStream);
      return -1;
    }

    // Make sure address is valid
    inst_id = inst_id - BACKEND_ADDRESS_START_SERIAL;
    if (inst_id < 0 || inst_id >= SERIAL_MAX_INSTRUMENTS) {
      fclose(routeFileStream);
      return -1;
    }

    // Save device to route table
    len = strlen(device);
    serial_routing_table_device[inst_id] = calloc(sizeof(char), len+1);
    if (!serial_routing_table_device[inst_id]) {
      fclose(routeFileStream);
      return -1;
    }
    strncpy(serial_routing_table_device[inst_id], device, len);

    // Save baudrate to route table
    serial_routing_table_baudrate[inst_id] = baudrate;

    // Save flowcontrol to route table
    if ( strcmp("none", flowcontrol) == 0 ) {
      serial_routing_table_flowcontrol[inst_id] = SERIAL_FLOWCONTROL_NONE;
    } else if ( strcmp("hardware", flowcontrol) == 0 ) {
      serial_routing_table_flowcontrol[inst_id] = SERIAL_FLOWCONTROL_HARDWARE;
    } else if ( strcmp("software", flowcontrol) == 0 ) {
      serial_routing_table_flowcontrol[inst_id] = SERIAL_FLOWCONTROL_SOFTWARE;
    } else {
      fclose(routeFileStream);
      return -1;
    }

  }

  fclose(routeFileStream);
  return 0;
} // end serial_load_inst_routes()


/*******************************************************************************
 * Set baudrate in newtio structure
 * Return 0 on success, -1 on failure
 ******************************************************************************/
int serial_inst_set_speed(int baudrate) {

  if (baudrate==0) {
    cfsetspeed(&newtio, B0);
  } else if (baudrate==50) {
    cfsetspeed(&newtio, B50);
  } else if (baudrate==75) {
    cfsetspeed(&newtio, B75);
  } else if (baudrate==110) {
    cfsetspeed(&newtio, B110);
  } else if (baudrate==134) {
    cfsetspeed(&newtio, B134);
  } else if (baudrate==150) {
    cfsetspeed(&newtio, B150);
  } else if (baudrate==200) {
    cfsetspeed(&newtio, B200);
  } else if (baudrate==300) {
    cfsetspeed(&newtio, B300);
  } else if (baudrate==600) {
    cfsetspeed(&newtio, B600);
  } else if (baudrate==1200) {
    cfsetspeed(&newtio, B1200);
  } else if (baudrate==1800) {
    cfsetspeed(&newtio, B1800);
  } else if (baudrate==2400) {
    cfsetspeed(&newtio, B2400);
  } else if (baudrate==4800) {
    cfsetspeed(&newtio, B4800);
  } else if (baudrate==9600) {
    cfsetspeed(&newtio, B9600);
  } else if (baudrate==19200) {
    cfsetspeed(&newtio, B19200);
  } else if (baudrate==38400) {
    cfsetspeed(&newtio, B38400);
  } else if (baudrate==57600) {
    cfsetspeed(&newtio, B57600);
  } else if (baudrate==115200) {
    cfsetspeed(&newtio, B115200);
  } else if (baudrate==230400) {
    cfsetspeed(&newtio, B230400);
  //} else if (baudrate==460800) {
  //  cfsetspeed(&newtio, B460800);
  } else {
    return -1;
  }

  return 0;

} // end serial_inst_set_speed()
