/*******************************************************************************
 * session.c
 * GPIB Session Server
 *
 * December 19, 2005
 *
 * Michael Siegenthaler
 * www.msigi.net
 *
 ******************************************************************************/
#define _GNU_SOURCE
#include <sys/types.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>

#include <signal.h>
#include <sys/wait.h>

#include <sys/socket.h>
#include <netinet/in.h>

#include "session.h"
#include "log.h"
#include "loadconfig.h"
#include "lionet.h"
#include "command.h"
#include "gpibcomm.h"
#include "tcpipcomm.h"
#include "serialcomm.h"

#define WELCOME_MESSAGE "NETGPIBD 1.1.0 C-Linux 2005-12-18"

#define RBUFLEN 1024

// Function declarations
void send_error(int errnum);
void send_ok(void);
void gpib_init_w(char* param);
void gpib_free_w(char* param);
void gpib_write_w(char* param);
void gpib_read_w(char* param);
void gpib_query_w(char* param);
void gpib_settimeout_w(char* param);
int get_instr_lock(int addr);
void release_instr_lock(void);

// Command set
command_t gpib_commands[] = {
  { "init", GC_INIT },
  { "write", GC_WRITE },
  { "read", GC_READ },
  { "query", GC_QUERY },
  { "free", GC_FREE },
  { "close", GC_CLOSE },
  { "ping", GC_PING },
  { "set timeout", GC_SET_TIMEOUT },
  { NULL, GC_INVALID }
};

// Flags
int initSuccess = 0;
//int alreadyInitializedBus = 0;

// Socket descriptor
int session_sd;

// Backend communication type (GPIB, Serial, TCP/IP)
int communication_type;

// GPIB instrument handle
int inst_fd = 0;

// Instrument lock
char* lockFileName = 0;

/*******************************************************************************
 * Main entry point for session
 ******************************************************************************/
void run_session(int session_sd_in) {

  // Variables
  char buf[RBUFLEN];
  int loop;
  parsed_command_t parsedCommand;

  // Store socket descriptor as a global variable for use by other functions
  session_sd = session_sd_in;

  // Tell the command parser about the valid set of commands
  set_command_set(gpib_commands);

  // Send welcome message
  send_line(session_sd, WELCOME_MESSAGE);

#ifdef LIONET_DEBUG
  while (1) {
    recv_line(session_sd, buf, RBUFLEN);
    send_line(session_sd, buf);
  }
#endif

  // Make sure to delete instrument lockfile on exit
  if (atexit(release_instr_lock) == -1) {
    logMessage("FATAL: Cannot register release_instr_lock() with atexit()");
    exit(EXIT_FAILURE);
  }

  // main loop for session
  while (1) {

    // Preparation
    loop = 1;
    initSuccess = 0;

    // Expect init, set, close, or ping
    while (loop) {

      // Get the command from the network
      recv_line(session_sd, buf, RBUFLEN);

      // Parse the command
      parsedCommand = parse_command(buf);
      switch (parsedCommand.cmdid) {
        case GC_INIT:
          // Initialize
          gpib_init_w(parsedCommand.param);
          if (initSuccess) { loop = 0; }
          break;
        case GC_SET_TIMEOUT:
          // Set timeout
          gpib_settimeout_w(parsedCommand.param);
          break;
        case GC_CLOSE:
          // End session
          send_ok();
          exit(EXIT_SUCCESS);
          break;
        case GC_PING:
          // Send reply and do nothing else
          send_line(session_sd, "pong");
          break;
        default:
          send_error(GE_BADCOMMAND);
      }

    }

    // Preparation
    loop = 1;

    // Expect write, read, query, free, close, or ping
    while (loop) {

      // Get the command from the network
      recv_line(session_sd, buf, RBUFLEN);

      // Parse the command
      parsedCommand = parse_command(buf);
      switch (parsedCommand.cmdid) {
        case GC_WRITE:
          // Write to instrument
          gpib_write_w(parsedCommand.param);
          break;
        case GC_READ:
          // Read from instrument
          gpib_read_w(parsedCommand.param);
          break;
        case GC_QUERY:
          // Query instrument
          gpib_query_w(parsedCommand.param);
          break;
        case GC_FREE:
          // Release instrument lock
          gpib_free_w(parsedCommand.param);
          loop = 0;
          break;
        case GC_CLOSE:
          // End session
          gpib_free_w(parsedCommand.param);
          exit(EXIT_SUCCESS);
          break;
        case GC_PING:
          // Send reply and do nothing else
          send_line(session_sd, "pong");
          break;
        default:
          send_error(GE_BADCOMMAND);
      }

    }

  }

}

/*******************************************************************************
 * Sends an error message
 ******************************************************************************/
void send_error(int errnum) {

  char* str = 0;
  char* msg;

  // FIX: this is neither elegant nor easily expandable
  if (errnum == GE_BADCOMMAND) {
    msg = "Bad command";
  } else if (errnum == GE_COMMERR) {
    msg = "Communications error";
  } else if (errnum == GE_FATALERR) {
    msg = "Fatal error";
  } else if (errnum == GE_RESOURCEUNAVAILABLE) {
    msg = "Resource unavailable";
  } else if (errnum == GE_INTERNALERROR) {
    msg = "Internal error";
  } else {
    msg = "Unknown error";
  }
 
  asprintf(&str, "error %d %s", errnum, msg);
  if (!str) {
    logMessage("FATAL: cannot allocate memory in send_error()");
    exit(EXIT_FAILURE);
  }
  send_line(session_sd, str);
  free(str);

}

/*******************************************************************************
 * Sends an ok message
 ******************************************************************************/
void send_ok() {
  send_line(session_sd, "ok");
}

/*******************************************************************************
 * Wrapper function for GPIB initializer
 ******************************************************************************/
void gpib_init_w(char* param) {

  // Variables
  int address;
  char * tcpip_config_file;
  char * serial_config_file;

  // Parse parameter to get a valid GPIB address
  if ( !param || ( sscanf(param, "%d", &address) != 1 ) ) {
    send_error(GE_BADCOMMAND);
    return;
  }

  // Make sure the requested instrument is not locked by another session
  if ( get_instr_lock(address) == -1 ) {
    send_error(GE_RESOURCEUNAVAILABLE);
    return;
  }

  // Determine backend communcation type
  if (address < BACKEND_ADDRESS_START_SERIAL) {
    communication_type = BACKEND_GPIB;
  } else if (address < BACKEND_ADDRESS_START_TCPIP) {
    communication_type = BACKEND_SERIAL;
  } else {
    communication_type = BACKEND_TCPIP;
  }

  // Initialize session to instrument
  if (communication_type == BACKEND_GPIB) {

    // Connect to instrument
    inst_fd = inst_init(address);
    if (inst_fd == -1) {
      release_instr_lock();
      send_error(GE_COMMERR);
      return;
    }

  } else if (communication_type == BACKEND_SERIAL) {

    // Load serial instrument routes
    serial_config_file = getParamByName("serialconfig");
    if (!tcpip_config_file) {
      logMessage("FATAL: serialconfig not specified in config");
      send_error(GE_INTERNALERROR);
      exit(EXIT_FAILURE);
    }
    if (serial_load_inst_routes(serial_config_file)) {
      logMessage("FATAL: failed to load serial instrument routes");
      send_error(GE_INTERNALERROR);
      exit(EXIT_FAILURE);
    }

    // Connect to instrument
    inst_fd = serial_inst_init(address);
    if (inst_fd == -1) {
      release_instr_lock();
      send_error(GE_COMMERR);
      return;
    }

  } else if (communication_type == BACKEND_TCPIP) {

    // Load tcpip instrument routes
    tcpip_config_file = getParamByName("tcpipconfig");
    if (!tcpip_config_file) {
      logMessage("FATAL: tcpipconfig not specified in config");
      send_error(GE_INTERNALERROR);
      exit(EXIT_FAILURE);
    }
    if (tcpip_load_inst_routes(tcpip_config_file)) {
      logMessage("FATAL: failed to load tcpip instrument routes");
      send_error(GE_INTERNALERROR);
      exit(EXIT_FAILURE);
    }

    // Connect to instrument
    inst_fd = tcpip_inst_init(address);
    if (inst_fd == -1) {
      release_instr_lock();
      send_error(GE_COMMERR);
      return;
    }

  }

  // On success, set a flag
  initSuccess = 1;
  send_ok();
}

/*******************************************************************************
 * Wrapper function for GPIB terminate session
 ******************************************************************************/
void gpib_free_w(char* param) {
  if (communication_type == BACKEND_GPIB) {
    inst_term(inst_fd);
  } else if (communication_type == BACKEND_SERIAL) {
    serial_inst_term(inst_fd);
  } else if (communication_type == BACKEND_TCPIP) {
    tcpip_inst_term(inst_fd);
  }
  release_instr_lock();
  send_ok();
}

/*******************************************************************************
 * Wrapper function for GPIB write
 ******************************************************************************/
void gpib_write_w(char* param) {
  if (param == 0) {
    send_error(GE_BADCOMMAND);
    return;
  }
  if (communication_type == BACKEND_GPIB) {
    if ( gpib_write(inst_fd, param) == -1 ) {
      send_error(GE_COMMERR);
      return;
    }
  } else if (communication_type == BACKEND_SERIAL) {
    if ( serial_inst_write(inst_fd, param) == -1 ) {
      send_error(GE_COMMERR);
      return;
    }
  } else if (communication_type == BACKEND_TCPIP) {
    if ( tcpip_inst_write(inst_fd, param) == -1 ) {
      send_error(GE_COMMERR);
      return;
    }
  }
  send_ok();
}

/*******************************************************************************
 * Wrapper function for GPIB read
 ******************************************************************************/
void gpib_read_w(char* param) {

  char buf[RBUFLEN];
  char* str;

  if (communication_type == BACKEND_GPIB) {
    if ( gpib_read(inst_fd, buf, RBUFLEN-1) == -1 ) {
      send_error(GE_COMMERR);
    } else {
      asprintf(&str, "read %s", buf);
      if (!str) {
        send_error(GE_INTERNALERROR);
        return;
      }
      send_line(session_sd, str);
      free(str);
    }
  } else if (communication_type == BACKEND_SERIAL) {
    if ( serial_inst_read(inst_fd, buf, RBUFLEN-1) == -1 ) {
      send_error(GE_COMMERR);
    } else {
      asprintf(&str, "read %s", buf);
      if (!str) {
        send_error(GE_INTERNALERROR);
        return;
      }
      send_line(session_sd, str);
      free(str);
    }
  } else if (communication_type == BACKEND_TCPIP) {
    if ( tcpip_inst_read(inst_fd, buf, RBUFLEN-1) == -1 ) {
      send_error(GE_COMMERR);
    } else {
      asprintf(&str, "read %s", buf);
      if (!str) {
        send_error(GE_INTERNALERROR);
        return;
      }
      send_line(session_sd, str);
      free(str);
    }
  }

}

/*******************************************************************************
 * Wrapper function for GPIB query
 ******************************************************************************/
void gpib_query_w(char* param) {

  // Variables
  char buf[RBUFLEN];
  char* str;

  // Write
  if (param == 0) {
    send_error(GE_BADCOMMAND);
    return;
  }
  if (communication_type == BACKEND_GPIB) {
    if ( gpib_write(inst_fd, param) == -1 ) {
      send_error(GE_COMMERR);
      return;
    }
  } else if (communication_type == BACKEND_SERIAL) {
    if ( serial_inst_write(inst_fd, param) == -1 ) {
      send_error(GE_COMMERR);
      return;
    }
  } else if (communication_type == BACKEND_TCPIP) {
    if ( tcpip_inst_write(inst_fd, param) == -1 ) {
      send_error(GE_COMMERR);
      return;
    }
  }

  // Read
  if (communication_type == BACKEND_GPIB) {
    if ( gpib_read(inst_fd, buf, RBUFLEN-1) == -1 ) {
      send_error(GE_COMMERR);
    } else {
      asprintf(&str, "read %s", buf);
      if (!str) {
        send_error(GE_INTERNALERROR);
        return;
      }
      send_line(session_sd, str);
      free(str);
    }
  } else if (communication_type == BACKEND_SERIAL) {
    if ( serial_inst_read(inst_fd, buf, RBUFLEN-1) == -1 ) {
      send_error(GE_COMMERR);
    } else {
      asprintf(&str, "read %s", buf);
      if (!str) {
        send_error(GE_INTERNALERROR);
        return;
      }
      send_line(session_sd, str);
      free(str);
    }
  } else if (communication_type == BACKEND_TCPIP) {
    if ( tcpip_inst_read(inst_fd, buf, RBUFLEN-1) == -1 ) {
      send_error(GE_COMMERR);
    } else {
      asprintf(&str, "read %s", buf);
      if (!str) {
        send_error(GE_INTERNALERROR);
        return;
      }
      send_line(session_sd, str);
      free(str);
    }
  }

}

/*******************************************************************************
 * Wrapper function for GPIB set timeout
 ******************************************************************************/
void gpib_settimeout_w(char* param) {

  int timeout_val;

  if (!param) {
    send_error(GE_BADCOMMAND);
    return;
  }

  if ( sscanf(param, "%d", &timeout_val) != 1 ) {
    send_error(GE_BADCOMMAND);
    return;
  }

  if (communication_type == BACKEND_GPIB) {
    if ( set_timeout_val(timeout_val) == -1 ) {
      send_error(GE_BADCOMMAND);
      return;
    }
  } else if (communication_type == BACKEND_SERIAL) {
    if ( serial_set_timeout_val(timeout_val) == -1 ) {
      send_error(GE_BADCOMMAND);
      return;
    }
  } else if (communication_type == BACKEND_TCPIP) {
    if ( tcpip_set_timeout_val(timeout_val) == -1 ) {
      send_error(GE_BADCOMMAND);
      return;
    }
  }

  send_ok();

}

/*******************************************************************************
 * Gets a lock on an instrument
 * Returns 0 on success, -1 on failure
 ******************************************************************************/
int get_instr_lock(int addr) {

  // Variables
  char* str;
  char* lockFileDir;
  int lockFileDirLen;
  char* lockFilePrefix;
  int lockFile_fd;

  // Figure out a name to use for the lockfile
  lockFileDir = getParamByName("lockdir");
  if (!lockFileDir) {
    logMessage("FATAL: lockdir not specified in config");
    send_error(GE_INTERNALERROR);
    exit(EXIT_FAILURE);
  }
  lockFilePrefix = getParamByName("lockprefix");
  if (!lockFilePrefix) {
    logMessage("FATAL: lockprefix not specified in config");
    send_error(GE_INTERNALERROR);
    exit(EXIT_FAILURE);
  }
  lockFileDirLen = strlen(lockFileDir);
  if (lockFileDir[lockFileDirLen-1] == '/') {
    lockFileDir[lockFileDirLen-1] = '\0'; // trim terminating slash
  }
  asprintf(&lockFileName, "%s/%s%d", lockFileDir, lockFilePrefix, addr);
  if (!lockFileName) {
    logMessage("FATAL: Cannot allocate memory for lockFileName");
    send_error(GE_INTERNALERROR);
    exit(EXIT_FAILURE);
  }
  
  // Try to create the lockfile
  lockFile_fd = open(lockFileName, O_WRONLY | O_CREAT | O_EXCL, 0644);
  if (lockFile_fd < 0) {
    free(lockFileName);
    lockFileName = 0;
    // Cannot create lockfile
    if (errno == EEXIST) {
      asprintf(&str, "Cannot obtain lock for instrument %d", addr);
      dLogMessage(str);
      return -1;
    } else {
      logMessage("FATAL: No permission to create lockfile");
      send_error(GE_INTERNALERROR);
      exit(EXIT_FAILURE);
    }
  }
  close(lockFile_fd);

  asprintf(&str, "Obtained lock for instrument %d", addr);
  dLogMessage(str);

  return 0;
}

/*******************************************************************************
 * Releases a lock on an instrument
 ******************************************************************************/
void release_instr_lock() {

  // Delete lock file, if it exists, and remove its name from memory
  if (lockFileName) {
    if ( unlink(lockFileName) < 0 ) {
      logMessage("ERROR: Cannot delete lockfile");
    }
    free(lockFileName);
    lockFileName = 0;
    logMessage("Released lock");
  }

}
