/*******************************************************************************
 * tcpipcomm.c
 * TCP/IP Instrument Driver Generic Functions
 *
 * December 18, 2005
 *
 * Michael Siegenthaler
 * www.msigi.net
 *
 ******************************************************************************/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>

#include <sys/types.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <netdb.h>

#include "tcpipcomm.h"
#include "session.h"
#include "lionet.h"
//#include "log.h"

int tcpip_timeout_val = 10; // Default timeout (in seconds)

// Routing table
char * tcpip_routing_table_host[TCPIP_MAX_INSTRUMENTS];
unsigned short int tcpip_routing_table_port[TCPIP_MAX_INSTRUMENTS];

/*******************************************************************************
 * Send an instrument command
 ******************************************************************************/
int tcpip_inst_write (int inst, char *cmd)
{

  int rv;
  rv = _send_line(inst, cmd);
  if (rv < 0) {
    return -1;
  }
  return 0;

} // end tcpip_inst_write()


/*******************************************************************************
 * Read from the instrument
 ******************************************************************************/
int tcpip_inst_read (int inst, char *buf, int buflen)
{

  int rv;
  rv = _recv_line_timeout(inst, tcpip_timeout_val, buf, buflen);
  if (rv != 0) return -1;
  return 0;

} // end tcpip_inst_read()

/*******************************************************************************
 * Initialize instrument
 * Returns an instrument handle on success
 ******************************************************************************/
int tcpip_inst_init (int inst_addr)
{
  int inst_sd, rv, inst_idx;

  struct hostent *he;
  struct sockaddr_in inst_ip_addr;

  // Make sure address is valid
  inst_idx = inst_addr - BACKEND_ADDRESS_START_TCPIP;
  if (inst_idx < 0 || inst_idx >= TCPIP_MAX_INSTRUMENTS) {
    return -1;
  }

  // Make sure we have a route to this instrument
  if (!tcpip_routing_table_host[inst_idx]) return -1;
  if (!tcpip_routing_table_port[inst_idx]) return -1;

  // Look up hostname
  if ((he=gethostbyname(tcpip_routing_table_host[inst_idx]))==NULL) {
    return -1;
  }

  // Create socket
  if ((inst_sd = socket(PF_INET, SOCK_STREAM, 0)) == -1) {
    return -1;
  }
 
  // Set up address structure
  inst_ip_addr.sin_family = AF_INET;
  inst_ip_addr.sin_port = htons(tcpip_routing_table_port[inst_idx]);
  inst_ip_addr.sin_addr = *((struct in_addr *)he->h_addr);
  memset(&(inst_ip_addr.sin_zero), '\0', 8);

  // Connect
  rv = connect(inst_sd, (struct sockaddr *)&inst_ip_addr, sizeof(struct sockaddr));
  if (rv == -1) {
    return -1;
  }

  return inst_sd;
} // end tcpip_inst_init()


/*******************************************************************************
 * Close communications channel to instrument
 ******************************************************************************/
int tcpip_inst_term (int inst)
{
  close(inst);
  return 0;
} // end tcp_inst_term()


/*******************************************************************************
 * Set timeout value
 ******************************************************************************/
int tcpip_set_timeout_val (int tmoutval)
{
  if (tmoutval < TCPIP_TIMEOUT_MIN || tmoutval > TCPIP_TIMEOUT_MAX) return -1;
  tcpip_timeout_val = tmoutval;
  return 0;
} // end tcpip_set_timout()

/*******************************************************************************
 * Load instrument routes
 * Return 0 on success, -1 on failure
 ******************************************************************************/
int tcpip_load_inst_routes (char * tcpip_config_file)
{

  FILE* routeFileStream = 0;
  char line[1024];
  unsigned int inst_id;
  unsigned int port;
  char hostname[600];
  int i;
  size_t len;

  for (i=0;i<TCPIP_MAX_INSTRUMENTS;i++) {
    tcpip_routing_table_host[i] = (char*)0;
    tcpip_routing_table_port[i] = (unsigned short int)0;
  }

  routeFileStream = fopen(tcpip_config_file, "r");
  if (!routeFileStream) {
    return -1;
  }

  while ( fgets(line, 1024, routeFileStream) != 0 ) {

    memset(hostname, 0, 600);

    if ( line[0] == '#' || line[0] == '\n' || line[0] == '\r' ) continue;
    if ( sscanf(line, "%d%599s%d", &inst_id, hostname, &port) != 3 ) {
      fclose(routeFileStream);
      return -1;
    }

    // Make sure address is valid
    inst_id = inst_id - BACKEND_ADDRESS_START_TCPIP;
    if (inst_id < 0 || inst_id >= TCPIP_MAX_INSTRUMENTS) {
      fclose(routeFileStream);
      return -1;
    }

    // Save hostname to route table
    len = strlen(hostname);
    tcpip_routing_table_host[inst_id] = calloc(sizeof(char), len+1);
    if (!tcpip_routing_table_host[inst_id]) {
      fclose(routeFileStream);
      return -1;
    }
    strncpy(tcpip_routing_table_host[inst_id], hostname, len);

    // Save port to route table
    tcpip_routing_table_port[inst_id] = (unsigned short int)port;

  }

  fclose(routeFileStream);
  return 0;
} // end tcpip_load_inst_routes()
